-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core for SIMON-JAMBU96 
--! @author     Tao Huang 
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       Modified based on the template written by Ekawat (ice) Homsirikamol
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use work.AEAD_pkg.all;



entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 48;   --! Data
        G_KEY_SIZE      : integer := 96;   --! Key
        G_TAG_SIZE      : integer := 48;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 3;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is

--! Registers
signal key_reg          : std_logic_vector(G_KEY_SIZE   -1 downto 0); -- register to store current key 
signal s_in             : std_logic_vector(G_DBLK_SIZE*2-1 downto 0); -- state used as I/O in block cipher, twice of the data block size 
signal sR               : std_logic_vector(G_DBLK_SIZE  -1 downto 0); -- state registers, represent (X,Y,R) in the paper 
signal bdo_r            : std_logic_vector(G_DBLK_SIZE  -1 downto 0); -- output register 

--! Signals
signal s_out            : std_logic_vector(G_DBLK_SIZE*2-1 downto 0); 

--! Controls
--!     Register
signal is_decrypt       : std_logic;
signal bdo_valid_r      : std_logic; 

--!     Combinatorial
signal b_start          : std_logic; 
signal b_done           : std_logic; 
signal b_ready          : std_logic; 
signal pre_proc         : std_logic; 
signal ld_key           : std_logic;
signal ld_npub          : std_logic;
signal update_state_ad  : std_logic;
signal update_state_pt  : std_logic;
signal decrypt_pt       : std_logic; 
signal omit_msg         : std_logic;  
signal update_final     : std_logic; 
signal gen_tag          : std_logic; 


--signal ld_npub          : std_logic;


type t_state is (S_INIT, S_WAIT_START, S_WAIT_KEY, 
        S_WAIT_NPUB, S_INIT_STATE, S_WAIT_INIT_DONE, S_PRE_PROCESS, S_PROC_AD0, S_PROC_AD1, S_PROC_AD2,
         S_PROC_PT0, S_PROC_PT1, S_PROC_PT2, S_PRE_FINAL, S_FINAL0, S_FINAL1, S_FINAL2,
        S_WAIT_TAG_AUTH);
signal state        : t_state;
signal nstate       : t_state;

function reverse_byte(aa: std_logic_vector) return std_logic_vector is
        variable bb : std_logic_vector(aa'high downto aa'low);
begin
    for i in ((aa'high+1)/8-1) downto 0 loop
        bb(i*8+7 downto i*8) := aa(aa'high-i*8 downto (aa'high+1)-8-i*8);
    end loop;
    return bb;
end function reverse_byte;

begin

--! =======================================================================
    --! Datapath
    --! =======================================================================
    process(clk)
    begin
        if rising_edge(clk) then

            bdo_valid_r <= '0';

            if (ld_key = '1') then
                key_reg  <= reverse_byte(key);
            end if;

            if (ld_npub = '1') then 
                s_in <= x"000000000000" & reverse_byte(bdi); -- init state set 
            elsif (pre_proc = '1') then 
                sR   <= s_out(95 downto 48); 
                s_in <= s_out xor x"000000000000000000000005"; 
            elsif (update_state_ad = '1') then 
                s_in(95 downto 48)      <= s_out(95 downto 48) xor reverse_byte(bdi); 
                s_in(47 downto  0)      <= s_out(47 downto  0) xor sR xor x"000000000001"; 
                sR                      <= sR xor s_out(95 downto 48) xor reverse_byte(bdi); 
            elsif (update_state_pt = '1') then 
                s_in(95 downto 48)      <= s_out(95 downto 48) xor reverse_byte(bdi); 
                s_in(47 downto  0)      <= s_out(47 downto  0) xor sR; 
                sR                      <= sR xor s_out(95 downto 48) xor reverse_byte(bdi); 
                if (omit_msg /= '1') then 
                    bdo_r   <= s_out(47 downto 0) xor 
                             sR xor reverse_byte(bdi); 
                    bdo_valid_r <= '1'; 
                else 
                    bdo_valid_r <= '0'; 
                end if;
            elsif (decrypt_pt = '1') then 
                
                if (omit_msg /= '1') then 
                    bdo_r <= s_out(47 downto 0) xor 
                             sR xor reverse_byte(bdi);  
                    bdo_valid_r <= '1'; 
                else 
                    bdo_valid_r <= '0';
                end if; 
                
                s_in(47 downto 0) <= s_out(47 downto 0) xor sR;
                if (bdi_eot = '1') then 
                    case (to_integer(unsigned(bdi_size))) is 
                        when 0 => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor x"000000000080";
                            sR <= sR xor s_out(95 downto 48) xor x"000000000080"; 
                        when 1 => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor (x"0000000080" & (bdi(47 downto 40) xor s_out(7 downto 0) xor sR(7 downto 0)));
                            sR <= sR xor s_out(95 downto 48) xor (x"0000000080" & (bdi(47 downto 40) xor s_out(7 downto 0) xor sR(7 downto 0)));
                        when 2 => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor ( x"00000080" & 
                                                  (bdi(39 downto 32) xor s_out(15 downto 8) xor sR(15 downto 8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto 0) xor sR(7  downto 0))  );                      
                            sR <= sR xor s_out(95 downto 48) xor ( x"00000080" & 
                                                  (bdi(39 downto 32) xor s_out(15 downto 8) xor sR(15 downto 8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto 0) xor sR(7  downto 0)) ); 
                        when 3 => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor ( x"000080" & 
                                                  (bdi(31 downto 24) xor s_out(23 downto 16) xor sR(23 downto 16)) &  
                                                  (bdi(39 downto 32) xor s_out(15 downto  8) xor sR(15 downto  8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto  0) xor sR(7  downto  0)) );                      
                            sR <= sR xor s_out(95 downto 48) xor ( x"000080" & 
                                                  (bdi(31 downto 24) xor s_out(23 downto 16) xor sR(23 downto 16)) &
                                                  (bdi(39 downto 32) xor s_out(15 downto  8) xor sR(15 downto  8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto  0) xor sR(7 downto   0)) ); 
                        when 4 => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor ( x"0080" & 
                                                  (bdi(23 downto 16) xor s_out(31 downto 24) xor sR(31 downto 24)) &
                                                  (bdi(31 downto 24) xor s_out(23 downto 16) xor sR(23 downto 16)) &  
                                                  (bdi(39 downto 32) xor s_out(15 downto  8) xor sR(15 downto  8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto  0) xor sR(7  downto  0)) );                      
                            sR <= sR xor s_out(95 downto 48) xor ( x"0080" & 
                                                  (bdi(23 downto 16) xor s_out(31 downto 24) xor sR(31 downto 24)) &
                                                  (bdi(31 downto 24) xor s_out(23 downto 16) xor sR(23 downto 16)) &
                                                  (bdi(39 downto 32) xor s_out(15 downto  8) xor sR(15 downto  8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto  0) xor sR(7  downto  0)) ); 
                        when 5 => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor ( x"80" & 
                                                  (bdi(15 downto  8) xor s_out(39 downto 32) xor sR(39 downto 32)) &
                                                  (bdi(23 downto 16) xor s_out(31 downto 24) xor sR(31 downto 24)) &
                                                  (bdi(31 downto 24) xor s_out(23 downto 16) xor sR(23 downto 16)) &  
                                                  (bdi(39 downto 32) xor s_out(15 downto  8) xor sR(15 downto  8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto  0) xor sR(7  downto  0)) );                      
                            sR <= sR xor s_out(95 downto 48) xor ( x"80" & 
                                                  (bdi(15 downto  8) xor s_out(39 downto 32) xor sR(39 downto 32)) &
                                                  (bdi(23 downto 16) xor s_out(31 downto 24) xor sR(31 downto 24)) &
                                                  (bdi(31 downto 24) xor s_out(23 downto 16) xor sR(23 downto 16)) &
                                                  (bdi(39 downto 32) xor s_out(15 downto  8) xor sR(15 downto  8)) &
                                                  (bdi(47 downto 40) xor s_out(7  downto  0) xor sR(7  downto  0)) ); 
                        when others => 
                            s_in(95 downto 48) <= s_out(95 downto 48) xor reverse_byte(bdi)  
                                            xor s_out(47 downto 0) xor sR;     
                            sR <= s_out(95 downto 48) xor reverse_byte(bdi) xor 
                                  s_out(47 downto 0);
                    end case; 
                else 
                    --! full block 
                    s_in(95 downto 48) <= s_out(95 downto 48) xor reverse_byte(bdi)  
                                            xor s_out(47 downto 0) xor sR;     
                    sR <= s_out(95 downto 48) xor reverse_byte(bdi) xor 
                          s_out(47 downto 0);
                end if; 
            elsif (update_final = '1') then 
                s_in(95 downto 48)      <= s_out(95 downto 48); 
                s_in(47 downto  0)      <= s_out(47 downto  0) xor sR xor x"000000000003"; 
                sR                      <= sR xor s_out(95 downto 48); 
            elsif (gen_tag = '1') then 
                bdo_r <= s_out(95 downto 48) xor s_out(47 downto 0) xor sR; 
                if (is_decrypt = '0') then 
                    bdo_valid_r <= '1'; 
                end if; 
            end if; 

        
        end if;
    end process;

    bdo         <= reverse_byte(bdo_r); 
    bdo_valid   <= bdo_valid_r; 
    msg_auth_valid <= '1' when bdo_r = reverse_byte(bdi) else '0';

    --! =======================================================================
    --! Control
    --! =======================================================================
    gSyncRst:
    if (not G_ASYNC_RSTN) generate
        process(clk)
        begin
            if rising_edge(clk) then
                if (rst = '1') then
                    state <= S_INIT;
                else
                    state <= nstate;
                end if;
            end if;
        end process;
    end generate;

    process(clk)
    begin
        if rising_edge(clk) then
            if (state = S_WAIT_START) then
                is_decrypt <= decrypt;
            end if;
        end if;
    end process;

    process(
        state, key_valid, key_update, is_decrypt, 
        bdi_valid, bdi_type, bdi_eot, bdi_eoi,
        bdo_ready, b_done)
    begin
        --! Internal
        nstate             <= state;
        ld_key             <= '0';
        ld_npub            <= '0';
        pre_proc           <= '0'; 
        update_state_ad    <= '0'; 
        update_state_pt    <= '0';
        decrypt_pt         <= '0'; 
        update_final       <= '0'; 
        gen_tag            <= '0'; 
        omit_msg           <= '0'; 

        --! External
        key_ready   <= '0';
        bdi_ready   <= '0';
        --bdo_valid   <= '0';
        msg_auth_done <= '0';
        b_start       <= '0'; 

        case state is
            when S_INIT =>
                nstate <= S_WAIT_START;

            when S_WAIT_START =>
                if (bdi_valid = '1') then
                    if (key_update = '1') then
                        nstate <= S_WAIT_KEY;
                    else
                        nstate <= S_WAIT_NPUB;
                    end if;
                end if;

            when S_WAIT_KEY =>
                if (key_valid = '1') then
                    key_ready <= '1';
                    ld_key    <= '1';
                    nstate    <= S_WAIT_NPUB;
                end if;

            when S_WAIT_NPUB =>
                bdi_ready <= '1';
                if (bdi_valid = '1') then
                    ld_npub <= '1';
                    nstate  <= S_INIT_STATE;
                end if; 

            when S_INIT_STATE => 
                b_start <= '1';
                nstate  <= S_WAIT_INIT_DONE; 

            when S_WAIT_INIT_DONE => 
                if (b_done = '1') then 
                    nstate <= S_PRE_PROCESS; 
                end if; 

            when S_PRE_PROCESS => 
                pre_proc <= '1'; 
                nstate <= S_PROC_AD0; 

            when S_PROC_AD0 => 
                b_start <= '1'; 
                nstate <= S_PROC_AD1; 

            when S_PROC_AD1 => 
                if (b_done = '1') then 
                    nstate <= S_PROC_AD2; 
                end if; 

            when S_PROC_AD2 => 
                if (bdi_valid = '1') then 
                    update_state_ad <= '1'; 
                    bdi_ready <= '1'; 
                    if (bdi_eot = '1') and (to_integer(unsigned(bdi_size)) < 8) then 
                        nstate <= S_PROC_PT0; 
                    else 
                        nstate <= S_PROC_AD0; 
                    end if; 
                end if; 

            when S_PROC_PT0 => 
                b_start <= '1'; 
                nstate <= S_PROC_PT1; 

            when S_PROC_PT1 => 
                if (b_done = '1') then 
                    nstate <= S_PROC_PT2;
                end if; 

            when S_PROC_PT2 => 
                if (bdi_valid = '1') then 
                    bdi_ready <= '1'; 
                    if (is_decrypt = '1') then 
                        decrypt_pt <= '1'; 
                    else 
                        update_state_pt <= '1'; 
                    end if; 
                    
                    if (bdi_eot = '1') and (to_integer(unsigned(bdi_size)) < 8) then 
                        if (to_integer(unsigned(bdi_size)) = 0) then 
                            omit_msg <= '1'; 
                        end if; 
                        nstate <= S_PRE_FINAL; 
                    else 
                        nstate <= S_PROC_PT0; 
                    end if; 
                end if;

            when S_PRE_FINAL => 
                b_start  <= '1';  
                nstate   <= S_FINAL0; 

            when S_FINAL0 => 
                if (b_done = '1') then 
                    update_final <= '1'; 
                    nstate       <= S_FINAL1; 
                end if; 

            when S_FINAL1 => 
                b_start <= '1';
                nstate  <= S_FINAL2; 

            when S_FINAL2 => 
                if (b_done = '1') then 
                    gen_tag <= '1'; 
                    if (is_decrypt = '1') then 
                        nstate <= S_WAIT_TAG_AUTH; 
                    else 
                        nstate <= S_INIT; 
                    end if; 
                end if; 
                   

            when S_WAIT_TAG_AUTH =>
                --! Wait to compare tag
                if (bdi_valid = '1') then                    
                    bdi_ready <= '1';
                    msg_auth_done <= '1';
                    nstate        <= S_INIT;
                end if; 

            when others =>
                null; 
        end case;

    end process;

    block_cipher: entity work.SIMON96
    port map (
        rst       => rst,
        clk       => clk,

        din       => s_in,
        dout      => s_out,
        key       => key_reg,

        start     => b_start,
        ready     => b_ready,
        done      => b_done
    );

end structure;